#include <classes/Layouter/GadTools.h>

#pragma -
#include <pragma/intuition_lib.h>
#pragma +

#include <string.h>

#define PLACETEXT_MASK (PLACETEXT_LEFT|PLACETEXT_RIGHT|PLACETEXT_ABOVE|PLACETEXT_BELOW|PLACETEXT_IN)
#define GTAUTO_SIZE LAYOUT_AUTOSIZE

static WORD getLabelWidth(struct GTGadgetC *g)
{
	IntuiText l;
	l.ITextFont = (struct TextAttr *) g->textAttr();
	l.IText = g->label();
	return l.IText ? IntuiTextLength(&l) : 0;
}

static WORD getLabelHeight(struct GTGadgetC *g)
{
	struct TextAttr *t = (struct TextAttr *) g->textAttr();
	return t ? t->ta_YSize : 0;
}

// *************************************************************

LGTButtonC::LGTButtonC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTButtonC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGButtonC()
{
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

LGTButtonC::LGTButtonC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTButtonC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGButtonC()
{
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

BOOL LGTButtonC::layoutWidth()
{
	if (autoWidth())
	{
		if (labelPlace() & (PLACETEXT_IN|PLACETEXT_ABOVE|PLACETEXT_BELOW))
		{
			setWidth(getLabelWidth(this)+innerBorderWidth);
		}
		else
			setWidth(innerBorderWidth + 20);
	};
	return TRUE;
}

BOOL LGTButtonC::layoutHeight()
{
	if (autoHeight())
	{
		if (labelPlace() & (PLACETEXT_IN|PLACETEXT_LEFT|PLACETEXT_RIGHT))
		{
			setHeight(getLabelHeight(this)+innerBorderHeight);
		}
		else
			setHeight(innerBorderHeight + 12);
	};
	return TRUE;
}

// *************************************************************

LGTStringC::LGTStringC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTStringC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGStringC()
{
	stringColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

LGTStringC::LGTStringC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTStringC(event,0,0,
		Width,Height, label,flags, tattr,id, (struct TagItem *) &tag1type),
	  LGStringC()
{
	stringColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

BOOL LGTStringC::layoutWidth()
{
	if (autoWidth())
	{
		IntuiText l;
		l.ITextFont = (struct TextAttr *) textAttr();
		l.IText = "0";
		ULONG i = 0;
		if (stringColumns == LAYOUT_AUTOSIZE)
		{
			i = taglist()->getData(GTST_MaxChars,20);
			if (i<4)
				i = 4;
			if (i>30)
				i = 30;
		}
		else
			i = stringColumns;
		setWidth(IntuiTextLength(&l)*i+innerBorderWidth);
	};
	return TRUE;
}

BOOL LGTStringC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(getLabelHeight(this)+innerBorderHeight);
	};
	return TRUE;
}

// *************************************************************

LGTIntegerC::LGTIntegerC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTIntegerC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGIntegerC()
{
	stringColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

LGTIntegerC::LGTIntegerC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTIntegerC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGIntegerC()
{
	stringColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

BOOL LGTIntegerC::layoutWidth()
{
	if (autoWidth())
	{
		IntuiText l;
		l.ITextFont = (struct TextAttr *) textAttr();
		l.IText = "0";
		ULONG i = 0;
		if (stringColumns == LAYOUT_AUTOSIZE)
		{
			i = taglist()->getData(GTST_MaxChars,20);
			if (i<2)
				i = 2;
			if (i>10)
				i = 10;
		}
		else
			i = stringColumns;
		setWidth(IntuiTextLength(&l)*i+innerBorderWidth);
	};
	return TRUE;
}

BOOL LGTIntegerC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(getLabelHeight(this)+innerBorderHeight);
	};
	return TRUE;
}

// *************************************************************

LGTCheckboxC::LGTCheckboxC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTCheckboxC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGCheckboxC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

LGTCheckboxC::LGTCheckboxC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTCheckboxC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGCheckboxC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
}

BOOL LGTCheckboxC::layoutWidth()
{
	if (autoWidth())
	{
		IntuiText l;
		l.ITextFont = (struct TextAttr *) textAttr();
		l.IText = "0";
		WORD w = IntuiTextLength(&l);
		if (w < 26)
			setWidth(26)
		else
			setWidth(w);
	};
	return TRUE;
}

BOOL LGTCheckboxC::layoutHeight()
{
	if (autoHeight())
	{
		WORD lh = getLabelHeight(this);
		if (lh > 11)
			setHeight(lh)
		else
			setHeight(11);
	};
	return TRUE;
}

// *************************************************************

LGTRadioC::LGTRadioC(GadgetEventC *event,
		stringarray &buttonlabels,
		ULONG place, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTRadioC(event,0,0,
		Width, Height, buttonlabels, label, flags, 0, tattr, id, taglist),
	  LGRadioC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTRadioC::LGTRadioC(GadgetEventC *event,
		stringarray &buttonlabels,
		ULONG place,STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTRadioC(event,0,0,
		Width, Height, buttonlabels, label, flags, 0, tattr, id,
		(struct TagItem *) &tag1type),
	  LGRadioC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTRadioC::layoutWidth()
{
	if (autoWidth())
	{
		setWidth(MX_WIDTH);
	};
	return TRUE;
}

BOOL LGTRadioC::layoutHeight()
{
	if (autoHeight())
	{
		STRPTR *labels = (STRPTR *) taglist()->getData(GTMX_Labels,NULL);
		if (labels)
		{
			LONG i = 0;
			while (*labels) {
				i++;
				labels++;
			};
			if (i > 0)
			{
				WORD lh = getLabelHeight(this);
				LONG oh = (lh > MX_HEIGHT) ? lh : MX_HEIGHT;
				setHeight(oh*i+(i-1)*taglist()->getData(GTMX_Spacing,0)+1);
			}
			else
				setHeight(MX_HEIGHT);
		}
		else
			setHeight(MX_HEIGHT);
	};
	return TRUE;
}

// *************************************************************

LGTCycleC::LGTCycleC(GadgetEventC *event,
		stringarray &cyclelabels,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTCycleC(event,0,0,
		Width, Height, cyclelabels, label, flags, tattr, id, taglist),
	  LGCycleC()
{
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTCycleC::LGTCycleC(GadgetEventC *event,
		stringarray &cyclelabels,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTCycleC(event,0,0,
		Width, Height, cyclelabels, label, flags, tattr, id,
		(struct TagItem *) &tag1type),
	  LGCycleC()
{
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTCycleC::layoutWidth()
{
	if (autoWidth())
	{
		STRPTR *labels = (STRPTR *) taglist()->getData(GTCY_Labels,NULL);
		if (labels)
		{
			WORD maxLabel = 0;
			IntuiText l;
			l.ITextFont = (struct TextAttr *) textAttr();
			while (*labels) {
				l.IText = *labels;
				WORD ll = IntuiTextLength(&l);
				if (ll > maxLabel)
					maxLabel = ll;
				labels++;
			};
			setWidth(maxLabel+20+innerBorderWidth);
		}
		else
			setWidth(20+innerBorderWidth);
	};
	return TRUE;
}

BOOL LGTCycleC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(getLabelHeight(this)+innerBorderHeight);
	};
	return TRUE;
}

// *************************************************************

LGTSliderC::LGTSliderC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTSliderC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGSliderC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTSliderC::LGTSliderC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTSliderC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGSliderC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTSliderC::layoutWidth()
{
	if (autoWidth())
	{
		setWidth(18);
	}
	return TRUE;
}

BOOL LGTSliderC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(12);
	}
	return TRUE;
}

// *************************************************************

LGTScrollerC::LGTScrollerC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTScrollerC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGScrollerC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTScrollerC::LGTScrollerC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTScrollerC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGScrollerC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTScrollerC::layoutWidth()
{
	if (autoWidth())
	{
		setWidth(18);
	}
	return TRUE;
}

BOOL LGTScrollerC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(12);
	}
	return TRUE;
}

// *************************************************************

LGTListviewC::LGTListviewC(GadgetEventC *event,
		WORD entries, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTListviewC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGListviewC()
{
	listEntries = entries;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTListviewC::LGTListviewC(GadgetEventC *event,
		WORD entries, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTListviewC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGListviewC()
{
	listEntries = entries;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTListviewC::layoutWidth()
{
	if (autoWidth())
	{
		WORD lw = getLabelWidth(this);
		if (lw > 120)
			setWidth(20 + lw)
		else
			setWidth(120);
	};
	return TRUE;
}

BOOL LGTListviewC::layoutHeight()
{
	if (autoHeight())
	{
		WORD lh = getLabelHeight(this);
		if (listEntries == LAYOUT_AUTOSIZE)
			setHeight(8*lh+4)
		else
			setHeight(listEntries*lh+4);
	};
	return TRUE;
}

// *************************************************************

LGTPaletteC::LGTPaletteC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTPaletteC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGPaletteC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTPaletteC::LGTPaletteC(GadgetEventC *event,
		STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTPaletteC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGPaletteC()
{
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTPaletteC::layoutWidth()
{
	if (autoWidth())
	{
		setWidth(100);
	};
	return TRUE;
}

BOOL LGTPaletteC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(100);
	};
	return TRUE;
}

// *************************************************************

LGTTextC::LGTTextC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTTextC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGTextC()
{
	textColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTTextC::LGTTextC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTTextC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGTextC()
{
	textColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTTextC::layoutWidth()
{
	if (autoWidth())
	{
		IntuiText l;
		l.ITextFont = (struct TextAttr *) textAttr();
		l.IText = "0";
		ULONG i = 0;
		if (textColumns == LAYOUT_AUTOSIZE)
		{
			i = 5;
			if (i<4)
				i = 4;
			if (i>30)
				i = 30;
		}
		else
			i = textColumns;
		setWidth(IntuiTextLength(&l)*i+innerBorderWidth);
	};
	return TRUE;
}

BOOL LGTTextC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(getLabelHeight(this)+innerBorderHeight);
	};
	return TRUE;
}

// *************************************************************

LGTNumberC::LGTNumberC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		struct TagItem *taglist)
	: GTNumberC(event,0,0,
		Width, Height, label, flags, tattr, id, taglist),
	  LGNumberC()
{
	textColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

LGTNumberC::LGTNumberC(GadgetEventC *event,
		WORD columns, STRPTR label, WORD Width, WORD Height,
		ULONG flags, const TextAttrC *tattr, UWORD id,
		Tag tag1type, ...)
	: GTNumberC(event,0,0,
		Width, Height, label, flags, tattr, id, (struct TagItem *) &tag1type),
	  LGNumberC()
{
	textColumns = columns;
	innerBorderWidth = 8;
	innerBorderHeight = 6;
	aWidth = (Width == GTAUTO_SIZE);
	aHeight = (Height == GTAUTO_SIZE);
};

BOOL LGTNumberC::layoutWidth()
{
	if (autoWidth())
	{
		IntuiText l;
		l.ITextFont = (struct TextAttr *) textAttr();
		l.IText = "0";
		ULONG i = 0;
		if (textColumns == LAYOUT_AUTOSIZE)
		{
			i = 20;
			if (i<4)
				i = 4;
			if (i>30)
				i = 30;
		}
		else
			i = textColumns;
		setWidth(IntuiTextLength(&l)*i+innerBorderWidth);
	};
	return TRUE;
}

BOOL LGTNumberC::layoutHeight()
{
	if (autoHeight())
	{
		setHeight(getLabelHeight(this)+innerBorderHeight);
	};
	return TRUE;
}

// *************************************************************
